use std::collections::HashMap;

use crate::schema::datum::Datum;
use crate::schema::entity::{Entity, EntityPart, EntityPartVisitor, EntityVisitor};

#[derive(Debug)]
pub enum PartType {
    Datum(&'static str),
    Assoc(&'static str),
}

#[derive(Debug)]
pub struct PartState {
    pub name: &'static str,
    pub ty: PartType,
}

impl PartState {
    fn build<EP: EntityPart>() -> Self {
        struct AssocCheck(Option<&'static str>);
        impl EntityVisitor for AssocCheck {
            fn visit<E: Entity>(&mut self) {
                self.0 = Some(E::entity_name());
            }
        }

        let mut acheck = AssocCheck(None);
        EP::Datum::accept_entity_visitor(&mut acheck);

        PartState {
            name: EP::part_name(),
            ty: match acheck.0 {
                Some(aname) => PartType::Assoc(aname),
                None => PartType::Datum(EP::Datum::sql_type()),
            },
        }
    }
}

#[derive(Debug)]
pub struct EntityState {
    pub name: &'static str,
    typeid: std::any::TypeId,

    pub parts: Vec<PartState>,
}

impl EntityState {
    fn build<E: Entity>() -> Self {
        #[derive(Default)]
        struct PartVisitor(Vec<PartState>);
        impl EntityPartVisitor for PartVisitor {
            fn visit<EP: EntityPart>(&mut self) {
                self.0.push(PartState::build::<EP>());
            }
        }

        let mut pv = PartVisitor::default();
        E::accept_part_visitor(&mut pv);

        Self {
            name: E::entity_name(),
            typeid: std::any::TypeId::of::<E>(),
            parts: pv.0,
        }
    }
}

#[derive(Default, Debug)]
pub struct EntityStateContainer {
    states: HashMap<&'static str, EntityState>,
}

impl EntityStateContainer {
    pub fn iter_states(&self) -> impl Iterator<Item = &EntityState> {
        self.states.values()
    }

    pub fn make_context(&mut self) -> EntityContext {
        EntityContext { container: self }
    }
}

pub struct EntityContext<'a> {
    container: &'a mut EntityStateContainer,
}

impl<'a> EntityVisitor for EntityContext<'a> {
    fn visit<E: Entity>(&mut self) {
        // three cases:
        // 1. we haven't seen this entity
        // 2. we've seen this entity before

        if self.container.states.contains_key(E::entity_name()) {
            return;
        }

        let entry = self.container.states.entry(E::entity_name());

        let entry = entry.or_insert_with(EntityState::build::<E>);
        // sanity-check
        if entry.typeid != std::any::TypeId::of::<E>() {
            panic!("Identical entity name but different typeid!");
        }

        struct RecursiveVisitor<'a, 'b>(&'a mut EntityContext<'b>);
        impl<'a, 'b> EntityPartVisitor for RecursiveVisitor<'a, 'b> {
            fn visit<EP: EntityPart>(&mut self) {
                EP::Datum::accept_entity_visitor(self.0);
            }
        }

        E::accept_part_visitor(&mut RecursiveVisitor(self));
    }
}
